{
  Copyright (c) 2020, 2021 Frank Fischer <frank-fischer@shadow-soft.de>

  This program is free software: you can redistribute it and/or
  modify it under the terms of the GNU General Public License as
  published by the Free Software Foundation, either version 3 of the
  License, or (at your option) any later version.

  This program is distributed in the hope that it will be useful, but
  WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
  General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program.  If not, see  <http://www.gnu.org/licenses/>
}

unit tests.csvreader;

{$mode objfpc}{$H+}
{$modeswitch typehelpers}

interface

uses fpcunit, testregistry, EasyCSV;

type
   TTestCSVReader = class(TTestCase)
   private
      Freader : TCSVReader;

   published
      procedure Test_ReadRow;
      procedure Test_Rows;
      procedure Test_ColumnBounds;
      procedure Test_Header;
      procedure Test_ReadNamedRow;
      procedure Test_NamedRows;
      procedure Test_NamedRows_Rename;
      procedure Test_Quoted;
      procedure Test_Whitespace;
      procedure Test_Delimiter;
      procedure Test_Integer;
      procedure Test_Integer10;
      procedure Test_Double;
      procedure Test_DoubleFormat;
      procedure Test_DefaultLineEndings;
      procedure Test_LineEnding;
      procedure Test_LineEndings;
      procedure Test_QuotedLineEndings;
      procedure Test_EnlargeBuffer;

   protected
      procedure Setup; override;
      procedure TearDown; override;

      class procedure AssertEquals(Expected : array of const; const Actual: array of String); overload;
   end;

implementation

uses SysUtils, Classes;

type
   TStringArrayHelper = type helper for TStringArray
      function Equals(other : TStringArray) : Boolean;
   end;

function TStringArrayHelper.Equals(other : TStringArray) : Boolean;
var
   i : Integer;
begin
   if Length(self) <> Length(other) then exit(false);
   for i := Low(self) to High(self) do
      if self[i] <> other[i] then exit(false);
   result := true;
end;

class procedure TTestCSVReader.AssertEquals(Expected : array of const; const Actual: array of String);
var
   i : Integer;
   actuals : TStringList = nil;
   expecteds : TStringList = nil;
begin
   try
      actuals := TStringList.Create;
      actuals.linebreak := ', ';
      actuals.skiplastlinebreak := true;
      expecteds := TStringList.Create;
      expecteds.linebreak := ', ';
      expecteds.skiplastlinebreak := true;

      actuals.AddStrings(actual);
      for i := 0 to High(expected) do
         case expected[i].vtype of
            vtinteger: expecteds.Add(expected[i].vinteger.tostring);
            vtboolean: expecteds.Add(expected[i].vboolean.tostring);
            vtchar: expecteds.Add(expected[i].vchar);
            vtstring: expecteds.Add(expected[i].vstring^);
            vtpchar: expecteds.Add(expected[i].vpchar);
            vtansistring: expecteds.Add(AnsiString(expected[i].vansistring));
         else
            Assert(False, 'Unsupported type');
         end;

      TTestCSVReader.AssertTrue('Expected: [' + expecteds.text + '], Actual: [' + actuals.text + ']', actuals.Equals(expecteds), CallerAddr);
   finally
      actuals.Free;
      expecteds.Free;
   end
end;

type
   TMyCSVReader = class(TCSVReader)
   public
      NextMaxBuffer: Integer;
      CntEnlarge: Integer;

   protected
      procedure EnlargeMaxBufferSize; override;

   public
      constructor Create;
   end;


procedure TTestCSVReader.Setup;
begin
   Freader := TCSVReader.Create;
end;

procedure TTestCSVReader.TearDown;
begin
   Freader.Free;
end;

procedure TTestCSVReader.Test_ReadRow;
var
   line : array of String;
begin
   Freader.SourceText := '1,2,3' + LineEnding + '4,5,6' + LineEnding + '7,8';

   AssertTrue(Freader.ReadRow(line));
   AssertEquals(['1', '2', '3'], line);

   AssertTrue(Freader.ReadRow(line));
   AssertEquals(['4', '5', '6'], line);

   AssertTrue(Freader.ReadRow(line));
   AssertEquals(['7', '8'], line);

   AssertFalse(Freader.ReadRow(line));

   AssertEquals(2, Freader.MinColumns);
   AssertEquals(3, Freader.MaxColumns);
end;

procedure TTestCSVReader.Test_Rows;
var
   row : TCSVReader.TRow;
   i : Integer = 0;
begin
   Freader.SourceText := '1,2,3' + LineEnding + '4,5,6' + LineEnding + '7,8';

   for row in Freader.Rows do begin
      inc(i);
      case i of
         1: begin
               AssertEquals('1', row[0]);
               AssertEquals('2', row[1]);
               AssertEquals('3', row[2]);
               AssertEquals(3, row.Count);
            end;
         2: begin
               AssertEquals('4', row[0]);
               AssertEquals('5', row[1]);
               AssertEquals('6', row[2]);
               AssertEquals(3, row.Count);
            end;
         3: begin
               AssertEquals('7', row[0]);
               AssertEquals('8', row[1]);
               AssertEquals(2, row.Count);
            end;
      else Assert(False);
      end
   end;

   AssertEquals(2, Freader.MinColumns);
   AssertEquals(3, Freader.MaxColumns);
end;

procedure TTestCSVReader.Test_ColumnBounds;
var
   line: TStringArray;
begin
   Freader.SourceText := '1,2,3' + LineEnding + '4,5,6' + LineEnding + '7,8';
   Freader.MinColumns := 3;

   AssertTrue(Freader.ReadRow(line));
   AssertTrue(Freader.ReadRow(line));
   try
      Freader.ReadRow(line);
      Fail('Expected read error');
   except on ECSVReadError do;
   end;

   Freader.SourceText := '1,2,3' + LineEnding + '4,5,6' + LineEnding + '7,8,9,10,11';
   Freader.MinColumns := 0;
   Freader.MaxColumns := 4;

   AssertTrue(Freader.ReadRow(line));
   AssertTrue(Freader.ReadRow(line));
   try
      Freader.ReadRow(line);
      Fail('Expected read error');
   except on ECSVReadError do;
   end;
end;

procedure TTestCSVReader.Test_Header;
var
   line : TStringArray;
   row : TCSVReader.TRow;
begin
   Freader.HasHeader := true;
   Freader.SourceText := 'A,B,C' + LineEnding + '1,2,3' + LineEnding + '4,5,6' + LineEnding + '7,8';

   AssertEquals(3, Freader.NumHeaders);
   AssertEquals('A', Freader.header[0]);
   AssertEquals('B', Freader.header[1]);
   AssertEquals('C', Freader.header[2]);

   AssertTrue(Freader.ReadRow(line));
   AssertEquals(['1', '2', '3'], line);

   Freader.SourceText := 'X,B,C' + LineEnding + '1,2,3' + LineEnding + '4,5,6' + LineEnding + '7,8';
   for row in Freader.Rows do;

   AssertEquals(3, Freader.NumRows);
   AssertEquals(3, Freader.NumHeaders);
   AssertEquals('X', Freader.header[0]);
   AssertEquals('B', Freader.header[1]);
   AssertEquals('C', Freader.header[2]);
   AssertEquals(-1, Freader.IndexOfHeader('A'));
   AssertEquals(0, Freader.IndexOfHeader('X'));
   AssertEquals(1, Freader.IndexOfHeader('B'));
   AssertEquals(2, Freader.IndexOfHeader('C'));
end;

procedure TTestCSVReader.Test_ReadNamedRow;
var
   line : array of String;
begin
   Freader.SourceText := 'C,B,A' + LineEnding + '1,2,3' + LineEnding + '4,5,6' + LineEnding + '7,8';

   Freader.NamedHeaders := ['A', 'C'];

   AssertTrue(Freader.ReadNamedRow(line));
   AssertEquals(['3', '1'], line);

   AssertTrue(Freader.ReadNamedRow(line));
   AssertEquals(['6', '4'], line);

   AssertTrue(Freader.ReadNamedRow(line));
   AssertEquals(['', '7'], line);

   AssertFalse(Freader.ReadNamedRow(line));

   AssertEquals(2, Freader.MinColumns);
   AssertEquals(3, Freader.MaxColumns);
end;

procedure TTestCSVReader.Test_NamedRows;
var
   row : TCSVReader.TRow;
   i : Integer = 0;
begin
   Freader.SourceText := 'C,B,A' + LineEnding + '1,2,3' + LineEnding + '4,5,6' + LineEnding + '7,8';

   for row in Freader.NamedRows(TStringArray.Create('A', 'C')) do begin
      inc(i);
      case i of
         1: begin
               AssertEquals('3', row[0]);
               AssertEquals('1', row[1]);
               AssertEquals(2, row.Count);
            end;
         2: begin
               AssertEquals('6', row[0]);
               AssertEquals('4', row[1]);
               AssertEquals(2, row.Count);
            end;
         3: begin
               AssertEquals('', row[0]);
               AssertEquals('7', row[1]);
               AssertEquals(2, row.Count);
            end;
      else Assert(False);
      end
   end;

   AssertEquals(2, Freader.MinColumns);
   AssertEquals(3, Freader.MaxColumns);
end;

procedure TTestCSVReader.Test_NamedRows_Rename;
var
   row : TCSVReader.TRow;
   i : Integer = 0;
begin
   Freader.SourceText := 'C,B,A' + LineEnding + '1,2,3' + LineEnding + '4,5,6' + LineEnding + '7,8';

   for row in Freader.NamedRows(TStringArray.Create('A', 'C')) do begin
      inc(i);
      case i of
         1: begin
               AssertEquals('3', row[0]);
               AssertEquals('1', row[1]);
               AssertEquals(2, row.Count);
               break;
            end;
      else Assert(False);
      end;
   end;

   for row in Freader.NamedRows(TStringArray.Create('B', 'A')) do begin
      inc(i);
      case i of
         2: begin
               AssertEquals('5', row[0]);
               AssertEquals('6', row[1]);
               AssertEquals(2, row.Count);
            end;
         3: begin
               AssertEquals('8', row[0]);
               AssertEquals('', row[1]);
               AssertEquals(2, row.Count);
            end;
      else Assert(False);
      end
   end;

   AssertEquals(2, Freader.MinColumns);
   AssertEquals(3, Freader.MaxColumns);
end;

procedure TTestCSVReader.Test_Quoted;
var
   line : TStringArray;
begin
   Freader.SourceText := '"1,2","2""3",3' + LineEnding + '4,5,6' + LineEnding + '7,8';
   AssertTrue(Freader.ReadRow(line));
   AssertEquals(['1,2', '2"3', '3'], line);
end;

procedure TTestCSVReader.Test_WhiteSpace;
const
   txt = '  1,2   ,   3   ' + LineEnding +
      ' "4"  ," 5 ", " 6  "  ' + LineEnding +
      '7,8';
var
   line : TStringArray;
begin
   Freader.SourceText := txt;
   AssertTrue(Freader.ReadRow(line));
   AssertEquals(['1', '2', '3'], line);
   AssertTrue(Freader.ReadRow(line));
   AssertEquals(['4', ' 5 ', ' 6  '], line);

   Freader.SourceText := txt;
   Freader.IgnoreOuterWhitespace := false;
   AssertTrue(Freader.ReadRow(line));
   AssertEquals(['  1', '2   ', '   3   '], line);
   AssertTrue(Freader.ReadRow(line));
   AssertEquals([' 4  ', ' 5 ', '  6    '], line);
end;

procedure TTestCSVReader.Test_Delimiter;
var
   row : TCSVReader.TRow;
begin
   Freader.SourceText := '1230456';
   Freader.Delimiter := '0';
   AssertTrue(Freader.ReadRow(row));
   AssertEquals(123, row.integers[0]);
   AssertEquals(456, row.integers[1]);
end;

procedure TTestCSVReader.Test_Integer;
var
   row : TCSVReader.TRow;
begin
   Freader.SourceText := '123,456' + LineEnding + '0789,-101112';

   AssertTrue(Freader.ReadRow(row));
   AssertEquals(123, row.integers[0]);
   AssertEquals(456, row.integers[1]);

   AssertTrue(Freader.ReadRow(row));
   AssertEquals(789, row.integers[0]);
   AssertEquals(-101112, row.integers[1]);
end;

procedure TTestCSVReader.Test_Integer10;
var
   row : TCSVReader.TRow;
   ok : Boolean;
begin
   Freader.SourceText := '$123' + LineEnding;
   AssertTrue(Freader.ReadRow(row));
   try
      ok := True;
      AssertEquals($123, row.integers[0]);
      ok := False;
   except ok := True end;
   AssertTrue(ok);

   Freader.SourceText := '&123' + LineEnding;
   AssertTrue(Freader.ReadRow(row));
   try
      ok := True;
      AssertEquals(&123, row.integers[0]);
      ok := False;
   except ok := True end;
   AssertTrue(ok);

   Freader.SourceText := '%101010' + LineEnding;
   AssertTrue(Freader.ReadRow(row));
   try
      ok := True;
      AssertEquals(%101010, row.integers[0]);
      ok := False;
   except ok := True end;
   AssertTrue(ok);

   Freader.SourceText := '0x123' + LineEnding;
   AssertTrue(Freader.ReadRow(row));
   try
      ok := True;
      AssertEquals($123, row.integers[0]);
      ok := False;
   except ok := True end;
   AssertTrue(ok);

   Freader.SourceText := 'x123' + LineEnding;
   AssertTrue(Freader.ReadRow(row));
   try
      ok := True;
      AssertEquals($123, row.integers[0]);
      ok := False;
   except ok := True end;
   AssertTrue(ok);

   Freader.SourceText := '0123' + LineEnding;
   AssertTrue(Freader.ReadRow(row));
   AssertEquals(123, row.integers[0]);

   Freader.SourceText := '1,,3' + LineEnding;
   AssertTrue(Freader.ReadRow(row));
   AssertEquals(1, row.integers[0]);
   AssertEquals(3, row.integers[2]);
   try
      ok := True;
      AssertEquals(0, row.integers[1]);
      ok := False;
   except ok := True end;
   AssertTrue(ok);
end;

procedure TTestCSVReader.Test_Double;
var
   row : TCSVReader.TRow;
begin
   Freader.SourceText := '12.3,.456,-12e23' + LineEnding + '07.89,-101.112,1e-3';

   AssertTrue(Freader.ReadRow(row));
   AssertEquals(12.3, row.doubles[0], 1e-9);
   AssertEquals(0.456, row.doubles[1], 1e-9);
   AssertEquals(-12e23, row.doubles[2], 1e-9);

   AssertTrue(Freader.ReadRow(row));
   AssertEquals(7.89, row.doubles[0]);
   AssertEquals(-101.112, row.doubles[1]);
   AssertEquals(1e-3, row.doubles[2]);
end;

procedure TTestCSVReader.Test_DoubleFormat;
var
   fmt: TFormatSettings;
   row : TCSVReader.TRow;
begin
   fmt := DefaultFormatSettings;
   fmt.DecimalSeparator := ',';

   Freader.Delimiter := ';';
   Freader.FormatSettings := fmt;
   Freader.SourceText := '12,3;,456;-12e23' + LineEnding + '07,89;-101,112;1e-3';

   AssertTrue(Freader.ReadRow(row));
   AssertEquals(12.3, row.doubles[0], 1e-9);
   AssertEquals(0.456, row.doubles[1], 1e-9);
   AssertEquals(-12e23, row.doubles[2], 1e-9);

   AssertTrue(Freader.ReadRow(row));
   AssertEquals(7.89, row.doubles[0]);
   AssertEquals(-101.112, row.doubles[1]);
   AssertEquals(1e-3, row.doubles[2]);
end;

procedure TTestCSVReader.Test_DefaultLineEndings;
var
   row : array of String;
begin
   Freader.SourceText := 'A,B,C'#10'1,2,3'#13'4,5,6'#10#13'7,8'#13#10'9,10';

   AssertTrue(Freader.ReadRow(row));
   AssertEquals(['A', 'B', 'C'], row);

   AssertTrue(Freader.ReadRow(row));
   AssertEquals(['1', '2', '3'], row);

   AssertTrue(Freader.ReadRow(row));
   AssertEquals(['4', '5', '6'], row);

   AssertTrue(Freader.ReadRow(row));
   AssertEquals(['7', '8'], row);

   AssertTrue(Freader.ReadRow(row));
   AssertEquals(['9', '10'], row);

   AssertFalse(Freader.ReadRow(row));
end;

procedure TTestCSVReader.Test_LineEnding;
var
   row : array of String;
begin
   Freader.SourceText := 'A,'#10',C\\\1,2,3\\\4,5,6\\\7,8\\\9,10';
   Freader.LineEnding := '\\\';

   AssertTrue(Freader.ReadRow(row));
   AssertEquals(['A', ''#10, 'C'], row);

   AssertTrue(Freader.ReadRow(row));
   AssertEquals(['1', '2', '3'], row);

   AssertTrue(Freader.ReadRow(row));
   AssertEquals(['4', '5', '6'], row);

   AssertTrue(Freader.ReadRow(row));
   AssertEquals(['7', '8'], row);

   AssertTrue(Freader.ReadRow(row));
   AssertEquals(['9', '10'], row);

   AssertFalse(Freader.ReadRow(row));
end;

procedure TTestCSVReader.Test_LineEndings;
var
   row : array of String;
begin
   Freader.SourceText := 'A,'#10',C\;1,2,3;\4,5,6\7,8;9,10';
   Freader.LineEndings := [';', '\', '\;', ';\'];

   AssertTrue(Freader.ReadRow(row));
   AssertEquals(['A', ''#10, 'C'], row);

   AssertTrue(Freader.ReadRow(row));
   AssertEquals(['1', '2', '3'], row);

   AssertTrue(Freader.ReadRow(row));
   AssertEquals(['4', '5', '6'], row);

   AssertTrue(Freader.ReadRow(row));
   AssertEquals(['7', '8'], row);

   AssertTrue(Freader.ReadRow(row));
   AssertEquals(['9', '10'], row);

   AssertFalse(Freader.ReadRow(row));
end;

procedure TTestCSVReader.Test_QuotedLineEndings;
var
   row : array of String;
begin
   Freader.SourceText := 'A,"B'#10#13'X",C'#10'1,2,3'#13'4,5,6'#10#13'7,8'#13#10'9,10';

   AssertTrue(Freader.ReadRow(row));
   AssertEquals(['A', 'B'#10#13'X', 'C'], row);

   AssertTrue(Freader.ReadRow(row));
   AssertEquals(['1', '2', '3'], row);

   AssertTrue(Freader.ReadRow(row));
   AssertEquals(['4', '5', '6'], row);

   AssertTrue(Freader.ReadRow(row));
   AssertEquals(['7', '8'], row);

   AssertTrue(Freader.ReadRow(row));
   AssertEquals(['9', '10'], row);

   AssertFalse(Freader.ReadRow(row));
end;

constructor TMyCSVReader.Create;
begin
   inherited;
   CntEnlarge := 0;
end;

procedure TMyCSVReader.EnlargeMaxBufferSize;
begin
   inc(CntEnlarge);
   if CntEnlarge = 1 then
      MaxBufferSize := 9
   else if CntEnlarge = 2 then
      MaxBufferSize := 15
   else
      MaxBufferSize := 25;
end;

procedure TTestCSVReader.Test_EnlargeBuffer;
var
   row : array of String;
begin
   Freader.Free;
   Freader := TMyCSVReader.Create;
   Freader.SourceText := 'short'#10'a,long,line'#10'a,very,very,long,line'#10;
   Freader.MinBufferSize := 2;
   Freader.MaxBufferSize := 2;

   AssertTrue(Freader.ReadRow(row));
   AssertEquals(['short'], row);

   AssertTrue(Freader.ReadRow(row));
   AssertEquals(['a', 'long', 'line'], row);

   AssertTrue(Freader.ReadRow(row));
   AssertEquals(['a', 'very', 'very', 'long', 'line'], row);

   AssertFalse(Freader.ReadRow(row));
   AssertEquals(3, TMyCSVReader(Freader).CntEnlarge);
end;

begin
   RegisterTest(TTestCSVReader);
end.
